//---------------------------------------------------------------------------

#ifndef deviceH
#define deviceH

#define INPUT_PORT 1
#define OUTPUT_PORT 2

// #include <vcl.h>
#include <list>
#include <set>
#include <map>
#include "UMStr.h"
#include "structure.h"


//---------------------------------------------------------------------------

class MV
{
  int count;
  int *matrix_not;
  int **matrix_and;

 public:
  int mNot(int val)
  {
    return matrix_not[val];
  }
  int mAnd(int left,int right)
  {
    return matrix_and[left][right];
  }
  int mOr(int left,int right)
  {
    return mNot(mAnd(mNot(left),mNot(right)));
  }
  int mXor(int left,int right)
  {
    return mOr(mAnd(mNot(left),right),mAnd(left,mNot(right)));
  }
  MV(int count,MStr m_not,MStr m_and):count(count)
  {
    string::iterator it=m_not.begin();
    matrix_not = new int[count];
    matrix_and = new int*[count];
    for(int i=0;i<count;i++)
    {
      matrix_not[i] = m_not[i] - '0';
      matrix_and[i] = new int[count];
      for(int j=0;j<count;j++)
        matrix_and[i][j]=m_and[j+count*i] - '0';
    }
  }
  MV(MV *mv):count(mv->count)
  {
    matrix_not = new int[mv->count];
    matrix_and = new int*[mv->count];
    for(int i=0;i<mv->count;i++)
    {
      matrix_not[i] = mv->mNot(i);
      matrix_and[i] = new int[mv->count];
      for(int j=0;j<mv->count;j++)
      {
        matrix_and[i][j] = mv->mAnd(i,j);
      }
    }
  }
  ~MV()
  {
    for(int i=0;i<count;i++)
      delete[] matrix_and[i];
    delete[] matrix_and;
    delete[] matrix_not;
  }
};

class Port;
class Wire;

class Object
{
 public:

   bool on_zjisteno;
   int id; // id in table of gates
   DDGate *myGate;
   int typobjektu;

   map<MStr,Wire *> wires;
   list<Port *> InputPorts;
   list<Port *> OutputPorts;
   map<MStr,Object *> objects;
   list<Object *> SortedObjects;

   MV *ValueType;
   void *AdditionalParameter1;
   void *AdditionalParameter2;

   Wire *find_wire(MStr name);
   Wire *addWire(MStr name);
   Port *getFirstInputPort();
   int ordinal_number;
   int fans; // number of fans of object
   MStr eval_param;
   MStr name;

  void eval();
  void eval(int);
  void evalMV(int);
  void evalPLA();

  int getLutEvalParam(MStr InitMStr);
  int getLutPortParam();


  void ClearBackedUps();
  void init(MStr);
  Object *copy();

  Object *findObject(const MStr &objectName);
  Port *findPort(const MStr &portName);
  Port *findOPort();
  Object *addObject(const MStr &ObjectName);
  Wire *addPRef(Port *,MStr); // pid ukazatel; pokud je teba, pid drt
  Wire *addPRef(Port *,Wire *); // pid ukazatel;
  Port *addPort(int flags,MStr name);
  void saveBench(MStr nof="");
  void saveCir(MStr nof="");
  void get_ports(int &inputs,int &outputs);

  int getOrdinalNumbers();

  void benchLikeDFFPorts();
  void benchLikeWireNames();
  bool evaluationDefined(const MStr&);


  Object(int count,MStr data_not,MStr data_and)
  {
    on_zjisteno=false;
    ValueType=new MV(count,data_not,data_and);
    AdditionalParameter1 = 0;
    AdditionalParameter2 = 0;
  }
  Object(MV *mv=0)
  {
    AdditionalParameter1 = 0;
    AdditionalParameter2 = 0;
    on_zjisteno=false;
    if(mv)
      ValueType=new MV(mv);
    else
      ValueType=0;
  }


  ~Object();
/*
  {
    if(eval_param == "PLA")
    {
      delete (list<pair<string,string> > *)AdditionalParameter1;
      delete[] (char *)AdditionalParameter2;
    }
    if(eval_param == "DFF")
    {
      delete (pair<bool,int> *)AdditionalParameter1;
    }
    delete ValueType;
    for(map<MStr,Object *>::iterator j=objects.begin();j!=objects.end();j++)
      delete j->second;
    for(list<Port *>::iterator i=InputPorts.begin();i!=InputPorts.end();i++)
      delete *i;
    for(list<Port *>::iterator i=OutputPorts.begin();i!=OutputPorts.end();i++)
      delete *i;
    for(map<MStr,Wire *>::iterator k=wires.begin();k!=wires.end();k++)
      delete k->second;
  }
  */
};

class Port
{
 public:
   int visited;
   int inoutflags;
   MStr name;
   Wire *wire_down;    // drt, kter vede z vnitku do portu
   Wire *wire_up;      // drt, kter vede ven z portu
   Object *object;
  Port(int io,MStr name,Object *o):inoutflags(io),name(name),object(o)
  {
    wire_up=0;
    wire_down=0;
  }
};

class Wire
{
 public:
   int visited;
   list<Port *> prefs;     //kam vude vede
   MStr name;
   int value;
   int OldValue;
   bool BackedUp;
   Wire(MStr name):name(name),BackedUp(false),value(false){}
};



//---------------------------------------------------------------------------
#endif
