/* Kamil Staufcik
 * with colleagues from
 * Czech Technical University in Prague
 * Faculty of Electrotechnics
 *
 */

//#include "edif.h"
#include "bench.h"
#include <math.h>
#include <unistd.h>
#include <getopt.h> // is required?
#include <time.h>
#include "UMStr.h"
#include "Upla.h"
#include "UBlif.h"
#include "structure.h"
#include "utility.h"

// #include "tables.h"

int dbg_print=1;
// char mujendl="\n";
char *mujendl;

int help(char *thisName) {
printf("Usage:\n%s [-ctejif] [-o output] [-s seed] [-v testfile] [-g debuglev] file\n",thisName);
printf("-c print out the circuit from outputs\n");
printf("-t set random test vector\n");
// printf("-e prints evaluation of tested vector\n");
printf("-j prints the injection\n");
printf("-i additional opening information\n");
printf("-f prints fault list\n");
printf("-o \"output\" saves input into file \"output\"\n");
printf("-s read random seed\n");
printf("-v read vectors from test vectors file\n");
printf("-g set debugging level to debuglev\n");
return 0;
}

//================================= main =====================================
int main(int argc,char **argv) {

 Object *o;
 DDObj *obvod;
 int check;

 mujendl=(char*)malloc(sizeof(char)*2);
 strcpy(mujendl,"");

// printf("%d - %s\n",argc,*argv); // print out params
	
// testovani objetu wire a gate
// testCirc();

// time_t cas; struktura pro cas - long int
// set up rnd generator
unsigned int seed=time(NULL);
srand(seed); 

// parsing params
int dbg_fault=0; // print the fault list
int dbg_open=0; // print info about opening
int dbg_circ1=0; // print circuit
int dbg_sim1=0; // evaluate vector
int dbg_inj1=0; // do injections
int dbg_setv1=0; // print vector
int dbg_testvf=0; // read vectors from file 
int dbg_level=0; // debugging info level 

char *outputF=NULL,*inputV=NULL;

int **testVectors=NULL;

int cl;
do {
cl=getopt(argc,argv,"ctejifo:s:v:g:");

switch (cl) {
	case -1: break;
	case 'c':
  		dbg_circ1=1; // print circuit
		 break;
	case 't':
  		dbg_setv1=1; // set random vector
		 break;
	case 'e':
  		dbg_sim1=1; // print evaluation
		 break;
	case 'j':
  		dbg_inj1=1; // injection
		 break;
	case 'i':
  		dbg_open=1; // additional information 
		 break;
	case 'f':
  		dbg_fault=1; // print fault list 
		 break;
	case 'g': // output file
		 if(optarg) {
			 dbg_level=atoi(optarg); }
		 break;
	case 's': // output file
		 if(optarg) {
			 seed=atol(optarg); }
		 break;
	case 'v': // output file
		 if(optarg) {
			 inputV=(char *)malloc((strlen(optarg)+1)*sizeof(char));
			 strcpy(inputV,optarg); }
			 dbg_testvf=1;
		 break;
	case 'o': // output file
		 if(optarg) {
			 outputF=(char *)malloc(strlen(optarg)+1);
			 strcpy(outputF,optarg); }
		 break;
	default: printf("Parametr je neplatny.\n"); help(argv[0]); exit(-1);
}
} while(cl!=-1);


int params=optind;

/*
if (optind < argc)  {
printf ("non-option ARGV-elements: ");
while (optind < argc)
   printf ("%s", argv[optind++]);
   printf ("\n");
}
//; exit (0);
*/


// do read some input

if (params<argc) 
{
  // read test vectors
  dbg_print=(dbg_level>1);
  if (dbg_testvf) testVectors=readVectors(inputV);

  dbg_print=((dbg_open==1)||(dbg_level>1)); // debug 

  // transform the structure
  obvod=transform(argv[params]);

  cout << "Random seed: " << seed << endl;

  // print out the circuit
  if (dbg_print=(dbg_circ1==1)) 
  	obvod->printCGO(); // print gates from outputs

  // ----------- fault list -----------
  // reducing fl
  /*
   DDFault *Ffault;
   DDGate *fromG;
   for(DDFault *xff=obvod->FaultList;xff!=NULL;xff=xff->next)
    if (xff->fg!=NULL) {
	     // printf("xff=%p prev=%p next=%p\n",xff,xff->prev,xff->next);
	    // printf("fault - i%d sa%d (gate %s - %p )\n",xff->index,xff->stuckAt,xff->fg->name,xff->fg);
	    fromG=xff->fg->inwires[xff->index]->fromgate;
	    if (fromG!=NULL) {
   	      // printf("from : %s - %p\n" ,fromG->name,fromG);
	        //  printf("xff=%p prev=%p Ff=%p\n",xff,xff->prev,Ffault);
		 Ffault=xff; xff=xff->prev; 
		 xff->next=Ffault->next;
		 if(xff->next!=NULL) xff->next->prev=xff;
	         // printf("xff=%p prev=%p Ff=%p\n",xff,xff->prev,Ffault);
		// delete Ffault;
	      // for(int i=0;i<fromG->inputs;i++) printf...
	    }
		   // else printf("from = null\n");
    }
   */
   DDFault *Ffault,**Efault;
   DDGate *fromG;
   for(DDFault *xff=obvod->FaultList;xff!=NULL;xff=xff->next)
    if (xff->fg!=NULL) {
	    
    }
  // print faults
     
  if (dbg_print=(dbg_fault==1)) 
   for(DDFault *xff=obvod->FaultList;xff!=NULL;xff=xff->next)
    if (xff->fg!=NULL) { 
	    if (xff->pio) printf("%s / %d (%d, PIN)\n",xff->fg->inwires[xff->index]->name,xff->stuckAt,xff->index); 
	    else printf("%s / %d (%d. at %s - %p)\n",xff->fg->inwires[xff->index]->name,xff->stuckAt,xff->index,xff->fg->name,xff->fg); }
 
 // ---------- fault injection --------- 
  if (dbg_print=(dbg_inj1==1)) 
  {
 // ---------- test vectors --------- 
  int *vvektor=makeArray<int>(obvod->inputs);
  do {
   if ((dbg_testvf)&&(testVectors!=NULL)) {
     if (testVectors[0]!=NULL) {
	vvektor=testVectors[0];
	testVectors++;
	// cout << "vv: " << testVectors[0] << endl;
     }
   }
   else {
  // create random input vector
   if (dbg_setv1==1) // make vector 
   for(int i=0;i<obvod->inputs;i++)
     vvektor[i]=rand()%2;  
   }
   dbg_print=(dbg_level>1);
   obvod->setVector(vvektor); // set vector to input
    
   dbg_print=(dbg_level>=1);
   // if (dbg_sim1==1)
   // cout << "Evaluate: " << endl;
   int *ov=obvod->Simulate(); // fsimulate 
   
   // if (dbg_sim1==1) {
   cout << "in: ";
   for (int i=0;i<obvod->inputs;i++) cout << vvektor[i] << "";
   cout << "-> out: ";
   for(int i=0;i<obvod->outputs;i++) cout << ov[i] << "";
   cout << endl;  
   // }

   // this is the place for real ATPG
   
   dbg_print=(dbg_level>=1);
   obvod->Inject(ov); // injection fxfsimulate
  
  } while ((testVectors!=NULL)&&(testVectors[0]!=NULL));
  }
//	printCircuit();
 /*
	cout << "vsechny poruchy: " << endl;
	faultList();

	cout << "redukovany seznam poruch: " << endl;
	faultListI();
*/
} // end arg 1
else { // no input file
  printf("Nothing to do...\n");
  help(argv[0]);
  return 0;
}

if (outputF) 
{
	printf("saving -  %s\n",outputF);
	save(o,argv[2]);
} // end arg 2

#ifdef devcpp
system("pause"); // do pause under win
#endif
}


//==end main==

