/* Kamil Staufcik
 * with colleagues from
 * Czech Technical University in Prague
 * Faculty of Electrotechnics
 */

//#include "edif.h"
#include "bench.h"
#include <math.h>
#include <unistd.h>
#include <getopt.h> // is required?
#include <time.h>
#include "UMStr.h"
#include "Upla.h"
#include "UBlif.h"
#include "structure.h"
#include "utility.h"

// #include "tables.h"

// parsing params
int dbg_fault=0; // print the fault list
int dbg_lfault=0; // print the full fault list
int dbg_open=0; // print info about opening
int dbg_circ1=0; // print circuit
int dbg_circ2=0; // print circuit
int dbg_sim1=0; // evaluate vector
int dbg_inj1=0; // do injections
int dbg_setv1=0; // print vector
int dbg_testvf=0; // read vectors from file 
int dbg_test=0; // print test vectors 
int dbg_flistf=0; // read fault list from file 
int dbg_masksf=0; // read masks from file 
int dbg_level=0; // debugging info level 
int dbg_ssim=0; // single pattern simulation
int dbg_reduce=0; // single pattern simulation

int dbg_generate=1; // do atpg
int dbg_basic=1; // basic information 
int dbg_print=1;

int max_vectors=10; // maximum of patterns/fault

// char mujendl="\n";
char *mujendl;

int help(char *thisName) {
printf("Usage:\n%s [-ctejif] [-o output] [-s seed] [-v testfile] [-g debuglev] file\n",thisName);
printf("-c print out the circuit from outputs\n");
printf("-b print out the circuit from inputs\n");
//printf("-t set random test vector\n");
printf("-s read random seed\n");
// printf("-e prints evaluation of tested vector\n");
printf("-o \"output\" saves input into bench file \"output\"\n");
printf("-g set debugging level to debuglev\n");
printf("-i additional opening information\n");
printf("-x do not print basic information\n");

printf("-f prints reduced fault list\n");
printf("-k read fault list from file\n");
printf("-l prints full fault list\n");
printf("-r fault list reduction\n");

printf("-V \"max. patterns\" to detect fault\n");

printf("-v read vectors from test vectors file\n");
printf("-m read mask vectors from mask file\n");
printf("-j do simulation using faults injection\n");
printf("-1 do single pattern simulation\n");
printf("\n");
printf("-y do not proceed atpg generation and simulation\n");
return 0;
}

//================================= main =====================================
int main(int argc,char **argv) {

 Object *o;
 DDObj *obvod;
 int check;

 mujendl=(char*)malloc(sizeof(char)*2);
 strcpy(mujendl,"");

// printf("%d - %s\n",argc,*argv); // print out params
	
// testovani objetu wire a gate
// testCirc();

// time_t cas; struktura pro cas - long int
// set up rnd generator
unsigned int seed=time(NULL);


char *outputF=NULL,*inputV=NULL,*maskV=NULL, *faultV=NULL;

int **testVectors=NULL;
int **maskVectors=NULL;

int cl;
do {
cl=getopt(argc,argv,"bctio:s:g:xv:rflk:m:yjeV:1");

switch (cl) {
	case -1: break;
	case 'b':
  		dbg_circ2=1; // print circuit from inputs
		 break;
	case 'c':
  		dbg_circ1=1; // print circuit from outputs
		 break;
	case 't':
  		// dbg_setv1=1; // set random vector
  		dbg_test=1; // print test vectors
		 break;
	case 'e':
  		dbg_sim1=1; // print evaluation
		 break;
	case 'o': // output file
		 if(optarg) {
			 outputF=(char *)malloc(strlen(optarg)+1);
			 strcpy(outputF,optarg); }
		 break;
	case 'i':
  		dbg_open=1; // additional information 
		 break;
	case 'x': // no basic info 
		 dbg_basic=0;
		 break;
	case 'g': // debug level
		 if(optarg) {
			 dbg_level=atoi(optarg); }
		 break;
	case 's': //  random seed
		 if(optarg) {
			 seed=atol(optarg); }
		 break;
	case 'f':
  		dbg_fault=1; // print fault list 
		 break;
	case 'l':
  		dbg_fault=1; // print fault list 
  		dbg_lfault=1; // print fault list 
		 break;
	case 'r':
  		dbg_reduce=1; // print fault list 
		 break;
	case '1': // single fault simulation 
		 dbg_ssim=1;
		 dbg_generate=0;
		 break;
	case 'y': // no test generation 
		 dbg_generate=0;
		 break;
	case 'V': // max. patterns 
		 if(optarg) {
			max_vectors=atol(optarg); }
		 break;
	case 'v': // test vectors input file 
		 if(optarg) {
			 inputV=(char *)malloc((strlen(optarg)+1)*sizeof(char));
			 strcpy(inputV,optarg); }
			 dbg_testvf=1;
			 dbg_generate=0;
		 break;
	case 'k': // fault list from file 
		 if(optarg) {
			 faultV=(char *)malloc((strlen(optarg)+1)*sizeof(char));
			 strcpy(faultV,optarg); }
			 dbg_flistf=1;
		 break;
	case 'm': // read masks from input file 
		 if(optarg) {
			 maskV=(char *)malloc((strlen(optarg)+1)*sizeof(char));
			 strcpy(maskV,optarg); }
			 dbg_masksf=1;
		 break;
	case 'j':
  		dbg_inj1=1; // injection
		 break;
	default: printf("Parametr je neplatny.\n"); help(argv[0]); exit(-1);
}
} while(cl!=-1);


int params=optind;

/* hint to read all non-opt args
if (optind < argc)  {
printf ("non-option ARGV-elements: ");
while (optind < argc)
   printf ("%s", argv[optind++]);
   printf ("\n");
}
*/

// do read some input

if (params<argc) 
{
  // read test vectors
  dbg_print=dbg_level;
  if (dbg_testvf) testVectors=readVectors(inputV);
  if (dbg_masksf) maskVectors=readVectors(maskV);

  dbg_print=((dbg_open==1)|(dbg_level)); // debug 
  // transform the structure
  if (((obvod=transform(argv[params]))==NULL)|| 
		  // error in structure or trasnform?
     (obvod->Outputs[0]==NULL))
  { // error in structure
	cerr << "Chyba ve vstupnim souboru!"  << endl;
	cerr << "(Obvod neni kombinacni?)"  << endl;
	DBG_Cout("obvod" << obvod << endl);
	DBG_Cout("output 0:" << obvod->Outputs[0] << endl);
	exit(-1);
  }

  /*  testing free function
  int a1=0,a2,b1=1,b2;
  DDGate *xg=obvod->Outputs[0]->fromgate;
  cout << xg->value << ":" << xg->svalue << endl;
  cout << obvod->Outputs[0]->fromgate->getFreeInput(-1,a1,a2,b1,b2) << endl;
  cout << a1 <<":" << a2 << ":" << b1 << ":" << b2 <<endl;
  a1=0;
  cout << obvod->Outputs[0]->fromgate->getFreeInput(0,a1,a2,b1,b2) << endl;
  cout << a1 <<":" << a2 << ":" << b1 << ":" << b2 <<endl;
  a1=0;
  cout << obvod->Outputs[0]->fromgate->getFreeInput(1,a1,a2,b1,b2) << endl;
  cout << a1 <<":" << a2 << ":" << b1 << ":" << b2 <<endl;
  */
  if (dbg_flistf) obvod->readFaults(faultV);

  if (dbg_basic) cerr << "Random seed: " << seed << endl;
  srand(seed); 

  // print out the circuit
  if (dbg_circ1) obvod->printCGO(); // print gates from outputs

  // print out the circuit from inputs
  if (dbg_circ2) obvod->printCGI(); // print gates from inputs

   dbg_print=dbg_level;

  // reduce faults
   if (dbg_reduce) reduceFL(obvod);

  // print faults
  if (dbg_fault) printFL(obvod,dbg_lfault); // print fault list, reduced r full -l 

  int *ov; // output vector

   // this is the place for real ATPG
  if (dbg_generate)  { // test generation
  	if (dbg_masksf) obvod->mask=maskVectors[0]; // !!! only 1st is now taken
	testVectors=makeTestVectors(obvod);
  	DBG_Cout("testV:" << testVectors << endl);
  }
  // --------- print vectors -------
  if ((dbg_generate||dbg_testvf)&&dbg_test) {
	// macro - some precompilers do not accept!
	/*
	 DBG_DO(
	 */
		for(int i=0;(i<obvod->wires*max_vectors)&&(testVectors[i]!=NULL);i++) {
		// cout << testVectors[i] << endl;
		printVector(testVectors[i]); }
	/*
		);
	*/
  }
 // ---------- fault injection --------- 
  if ((dbg_generate||dbg_testvf)&&(dbg_inj1))
  {
 // ---------- test vectors --------- 
  int *vvektor=makeArray<int>(obvod->inputs);
  int det=0;
  int fno=0;
  int undet=0;
  int totfl=0;
  do {
   // if ((dbg_testvf)&&(testVectors!=NULL)) {
   if (testVectors!=NULL) {
     if (testVectors[0]!=NULL) {
	vvektor=testVectors[0];
	testVectors++;
	// cout << "vv: " << testVectors[0] << endl;
     }
   }
   /*
   else {
  // create random input vector
   if (dbg_setv1==1) // make vector 
   for(int i=0;i<obvod->inputs;i++)
     vvektor[i]=rand()%2;  
   }
   dbg_print=(dbg_level>1);
   */ 
   if (obvod->setVector(vvektor)) { // set vector to input - if possible
   	// if (dbg_sim1==1)
	   // cout << "Evaluate: " << endl;
	   ov=obvod->Simulate(); // fsimulate 
   
	   // if (dbg_sim1==1) {
	   cout << "no fault: ";
	   printVector(vvektor,obvod->inputs);
	   cout << " ";
	   printVector(ov,obvod->outputs);
	   cout << endl;  
	   // }
	   fno++;
	   if (obvod->InjectX(ov)) { // injection fxfsimulate
	   det++;
	   }
   } else { // if setV...
   	cout << "mask mismatch: " << endl;
   	printVector(vvektor,obvod->inputs);
	cout << endl;
   	printVector(obvod->mask,obvod->inputs);
	cout << endl;
   	obvod->setVector(obvod->mask); // clear changes
   }
   

  
  } while ((testVectors!=NULL)&&(testVectors[0]!=NULL));
  DDFault *pom=obvod->FaultList;
  while (pom!=NULL) {
	  if (pom->active==1) undet++;
	  pom=pom->next;
	  totfl++;
  }
  cout << "Total faults: " << totfl << endl;
  cout << "Total tested: " << fno << "             " << (100.0*fno/totfl) << endl;
  cout << "Total detected: " << det << "              " << endl;
  cout << "Total undetected: " << undet << endl;
 } // fault injection

//	printCircuit();
 /*
	cout << "vsechny poruchy: " << endl;
	faultList();

	cout << "redukovany seznam poruch: " << endl;
	faultListI();
*/
  if (dbg_ssim)  {
  	// printf("Single simulation\n");
	while(!obvod->FaultList->isEmpty()) {
		testVectors=makeTestVectorsS(obvod);
		if ((testVectors!=NULL)&&(testVectors[0]!=NULL)) {
   		if (obvod->setVector(testVectors[0])) { // set vector to input
   			ov=obvod->Simulate(); // fsimulate 
 	  		obvod->InjectX(ov); // injection fxfsimulate
		} else { // mask mismatch
			obvod->setVector(obvod->mask);
		}

		}
	}
  DDFault *pom=obvod->FaultList;
  int totfl=0,fno=0,det=0,undet=0;
  while (pom!=NULL) {
	  if (pom->active==1) undet++;
	  pom=pom->next;
	  totfl++;
  }
  cout << "Total faults: " << totfl << endl;
  cout << "Total tested: " << fno << "             " << (100.0*fno/totfl) << endl;
  cout << "Total detected: " << det << "              " << endl;
  cout << "Total undetected: " << undet << endl;
  }
} // end arg 1 - params
else { // no input file
  printf("Nothing to do...\n");
  help(argv[0]);
  return 0;
}

if (outputF) 
{
	printf("saving -  %s\n",outputF);
	save(o,argv[2]);
} // end arg 2

#ifdef devcpp
system("pause"); // do pause under win
#endif
}


//==end main==

