/**********************************
* Structure of net 
*
**********************************/

#ifndef hStructure
#define hStructure

#include <stdio.h>
#include <stdlib.h>
#include <string>
#include "UMStr.h"
#include "logic4.h"
#include "debug.h"


#define tAND 1
#define tOR 2
#define tNOT 4
#define tXOR 8
#define tMUX 16
#define tUND 128 
#define tOBJ 256 

#define tNAND 5
#define tNOR 6
#define tXNOR 12 
#define tNMUX 20 

#define pUndef 0
#define pOpen 1
#define pBack 2
#define pClosed 3
#define pSens 4

#define vUndef -1

#define errMem 7777;

// sablona pro vytvareni ruznych poli
template<typename _Typ>
_Typ* makeArray(int i) {
	_Typ *memP;
     //if(i>0) { // allocate nonzero mem?
	memP=(_Typ *)(malloc(i*sizeof(_Typ)));
	if (memP<=0) throw errMem;
     /*
     } else // null if zero len
	memP=NULL;
     */
	for(int j=0;j<i;j++) memP[j]=(_Typ)(0);
	// for(int j=0;j<i;j++) memP[j]=(_Typ)(NULL);
	return memP; 
}

//----------- globalni promenne -------------

extern int dbg_print;
extern int dbg_basic;
extern char *mujendl;

//----------- trida vodicu-------------
class DDWire;
class DDFault;

//----------- trida hradel-------------
class DDGate {
public:
int gTyp;
int level,olevel;
int state; // opened/closed/computing...
int inputs; // outputs always 1!
int connected; // already connected wires 
int un_checked;
int value,svalue,vmask; // actual value and dont'care
char *name;
DDWire **inwires, *outwire; 
DDGate *backG;

void DDClear();
int getFreeInput(int i,int &nval,int &otval,int &sval,int &otsval);
DDGate() { DDClear(); inputs=2; inwires=makeArray<DDWire *>(3);  };
DDGate(int typ,int ins,int lev,const char *nam) {  
			DDClear();
			gTyp=typ; 
			level=lev;
			name=strdup(nam);
			inputs=ins;  
			inwires=makeArray<DDWire *>(ins+1); 
			outwire=NULL;
};
~DDGate() { };
};

//----------- trida vodicu-------------
class DDWire {
public:
int cpin;
int outputs; // input always 1
int connected; // already connected wires 
int un_checked; //
int state;
int level,olevel;
int value,svalue,vmask; // actual value and don't care
char *name; // wire name
DDFault *FL;
logic4Set *PINs;

DDGate *fromgate, *togate; // odkud, kam 
DDWire *fromwire, **towires;
void DDClear(); // inicializace
void makeLSet(int bits);
DDWire() { DDClear(); } // - drat bez inicializace
DDWire(const char *nam); // - drat se jmenem
DDWire(DDGate *in,DDGate *out,int whto,const char *nam); // - drat od hradla ke hradlu
DDWire(int outs,const char *nam,int mainins); // n-fan
~DDWire() {};
};

class DDItem {
public:
	DDItem **Inputs;
	DDItem **Outputs;
	DDFault *next,*prev; // next&prev item
	int inputs,outputs;
	DDItem() { inputs=0; outputs=0; };
	DDItem(int in,int out);
};

class DDFault:public DDItem {
public:
	DDGate *fg; // faulty gate
	DDWire *fw; // faulty wire 
	DDFault *ekvf; // next equivalent fault
	DDFault *na; // next active fault
	DDFault **domine,**domini; // dominating and dominated faults 
	int stuckAt,index,pio,active,found; // stuck at, index of wire
	void DDClear ();
	DDFault () { this->DDClear(); };
	DDFault(DDFault *pr,int sa,DDWire *flw);
	DDFault(DDFault *pr,int sa,DDGate *flg,int ind);
        int setPIO() { pio=1; }
	int isEmpty();
	int compareFS(char *);
	DDFault *getActive();
       ~DDFault() {};
};

class DDObj {
public:
	DDFault *FaultList;
	DDWire **Inputs,**Outputs;
	int *mask;
	int inputs,outputs;
	int un_checked;
	int wires;
DDObj(DDFault *FL,DDWire **IWs,DDWire **OWs,int ins,int outs) {
	FaultList=FL;
	Inputs=IWs;
	Outputs=OWs;
	inputs=ins;
	outputs=outs;
	un_checked=0;
	wires=vUndef;
	}

void printCGI();
void printCFI();
void printCGO();
void printCFO();
void Recalc(); // recalculate levels

int setVector(int *vvektor);
void readFaults(char *fname);
int *Simulate();
void Inject(int *correct);
int  InjectX(int *correct);

~DDObj() {};
};

extern void printcircg(DDGate *D,int uncheck);
extern void printwireg(DDWire *W,int uncheck);

extern void printcirc(DDGate *D);
extern void printwire(DDWire *W);

extern void connectGoutW(DDGate *G,DDWire *W);
extern void connectGinW(DDGate *G,DDWire **W);
extern void connectWfromG(DDWire *W,DDGate *G);
extern void connectWtoG(DDWire *W,DDGate **G);
extern void connectWtoGn(DDWire *W,DDGate *G,int m,int n);

extern void printback(DDGate *D);
extern void printbackw(DDWire *W);

extern void printbackg(DDGate *D,int uncheck);
extern void printbackwg(DDWire *W,int uncheck);

extern int simulate(DDGate *DG,int &check);
extern int simulatew(DDWire *WG,int &check);

// extern int propagate(DDGate *DG,int input, int forceValue, DDGate **NotSure,int &check);
extern int propagate(DDGate *DG,int input, int forceValue, DDGate **NotSure,DDGate **FanOuts,int &check);

extern int set_up(DDGate *DG, int forceValue, DDGate **NotSure,int &check);

extern const char *getGName(int gTyp);
extern int getType(MStr eval_param);

#endif
