/**********************************
* Structure of net 
*
**********************************/

#ifndef hStructure
#define hStructure

#include <stdio.h>
#include <stdlib.h>
#include "UMStr.h"
#include <string>

#define tAND 1
#define tOR 2
#define tNOT 4
#define tXOR 8
#define tMUX 16
#define tUND 128 
#define tOBJ 256 

#define pUndef 0
#define pOpen 1
#define pBack 2
#define pClosed 3

#define errMem 7777;


// sablona pro vytvareni ruznych poli
template<typename _Typ>
_Typ* makeArray(int i) {
	_Typ *memP;
     //if(i>0) { // allocate nonzero mem?
	memP=(_Typ *)(malloc(i*sizeof(_Typ)));
	if (memP<=0) throw errMem;
     /*
     } else // null if zero len
	memP=NULL;
     */
	for(int j=0;j<i;j++) memP[j]=(_Typ)(0);
	// for(int j=0;j<i;j++) memP[j]=(_Typ)(NULL);
	return memP; 
}

//----------- globalni promenne -------------

extern int dbg_print;
extern char *mujendl;

//----------- trida vodicu-------------
class DDWire;
class DDFault;

//----------- trida hradel-------------
class DDGate {
public:
int gTyp;
int level;
int state; // opened/closed/computing...
int inputs; // outputs always 1!
int connected; // already connected wires 
int un_checked;
int value; // actual value
char *name;
DDFault *FL;
DDWire **inwires, *outwire; 

DDGate() { inputs=2; inwires=makeArray<DDWire *>(3); outwire=NULL; gTyp=tUND; FL=NULL; };
DDGate(int typ,int ins,int lev,const char *nam) {  
			gTyp=typ; 
			level=lev;
			name=strdup(nam);
			inputs=ins;  
			inwires=makeArray<DDWire *>(ins+1); 
			outwire=NULL;
			FL=NULL;
};
~DDGate() { };
};

//----------- trida vodicu-------------
class DDWire {
public:
int outputs; // input always 1
int connected; // already connected wires 
int level; 
int value; // actual value
char *name; // wire name

DDGate *fromgate, **togates; // odkud, kam 
// - drat bez inicializace
DDWire() { int outputs=1; fromgate=NULL; togates=makeArray<DDGate *>(2); name=strdup("-");  };
// - drat se jmenem
DDWire(const char *nam) { outputs=1; fromgate=NULL; togates=makeArray<DDGate *>(2); name=strdup("-"); name=strdup(nam); };

// - drat od hradla ke hradlu
DDWire(DDGate *in,DDGate *out,int whto,const char *nam) { outputs=1; togates=&out; fromgate=in; out->inwires[whto]=this; in->outwire=this; name=strdup(nam); }
// - drat s vetvenim od 1 do dvou hradel
DDWire(DDGate *in,DDGate *out1, DDGate *out2, const char *nam) { outputs=2; fromgate=in; togates=makeArray<DDGate *>(3); togates[0]=out1; togates[1]=out2; togates[2]=in; name=strdup(nam); };

DDWire(int outs,const char *nam) { 
				outputs=outs; 
				name=strdup(nam);
				togates=makeArray<DDGate *>(outs+1); 
				fromgate=NULL;
};
~DDWire() {};
};

class DDFault {
public:
	DDGate *fg; // faulty gate
	DDFault *next,*prev; // next&prev fault
	DDFault *ekvf; // next equivalent fault
	int stuckAt,index,pio,active; // stuck at, index of wire
	DDFault () { 
		next=NULL; prev=NULL; fg=NULL; stuckAt=-1; index=-1; pio=0; active=1;
	};
	DDFault(DDFault *pr,int sa,DDGate *flg,int ind) { 
	    	if (dbg_print) 
			printf("novy fault, i=%d sa=%d fg=%p\n",ind,sa,flg);
      		stuckAt=sa; index=ind; fg=flg; pio=0;
		active=1;
      		prev=pr; 
		if (pr==NULL) { cerr << "vnitrni chyba\n"; exit(1); }
		if (dbg_print) cout<< "next\n"; 
      		next=pr->next; 
		if (next!=NULL) next->prev=this;
      		pr->next=this; 
      		if (sa==0) next=new DDFault(this,1,flg,ind); 
        }
        int setPIO() { pio=1; }
       ~DDFault() {};
};

class DDObj {
public:
	DDFault *FaultList;
	DDWire **Inputs,**Outputs;
	int inputs,outputs;
	int un_checked;
DDObj(DDFault *FL,DDWire **IWs,DDWire **OWs,int ins,int outs) {
	FaultList=FL;
	Inputs=IWs;
	Outputs=OWs;
	inputs=ins;
	outputs=outs;
	un_checked=0;
	}

void printCGI();
void printCFI();
void printCGO();
void printCFO();

void setVector(int *vvektor);
int *Simulate();
void Inject(int *correct);

~DDObj() {};
};

extern void printcircg(DDGate *D,int uncheck);
extern void printwireg(DDWire *W,int uncheck);

extern void printcirc(DDGate *D);
extern void printwire(DDWire *W);

extern void connectGoutW(DDGate *G,DDWire *W);
extern void connectGinW(DDGate *G,DDWire **W);
extern void connectWfromG(DDWire *W,DDGate *G);
extern void connectWtoG(DDWire *W,DDGate **G);
extern void connectWtoGn(DDWire *W,DDGate *G,int m,int n);

extern void printback(DDGate *D);
extern void printbackw(DDWire *W);

extern void printbackg(DDGate *D,int uncheck);
extern void printbackwg(DDWire *W,int uncheck);

extern int simulate(DDGate *DG,int &check);
extern int simulatew(DDWire *WG,int &check);

// extern int propagate(DDGate *DG,int input, int forceValue, DDGate **NotSure,int &check);
extern int propagate(DDGate *DG,int input, int forceValue, DDGate **NotSure,DDGate **FanOuts,int &check);

extern int set_up(DDGate *DG, int forceValue, DDGate **NotSure,int &check);

extern const char *getGName(int gTyp);
extern int getType(MStr eval_param);

#endif
